--[[
	MyRolePlay 4 (C) 2010-2019 Etarna Moonshyne <etarna@moonshyne.org>, Katorie @ Moon Guard
	Licensed under GNU General Public Licence version 2 or, at your option, any later version

	Utilities.lua - Basic utility functions used throughout the addon. Will be shifting commonly used ones to this over time.
]]

-- Return an texture text tag based on the given icon url and size. Nil safe.
function mrp:ReturnTextureTag(iconPath, iconSize)
	iconPath = iconPath or "INV_MISC_QUESTIONMARK";
	iconSize = iconSize or 15;
	return "|T" .. iconPath .. ":" .. iconSize .. ":" .. iconSize .. "|t"
end



--- IMAGE_PATTERN is the string pattern used for performing image replacements
--  in strings that should be rendered as HTML.
---
--- The accepted form this is "{img:<src>:<width>:<height>[:align]}".
---
--- Each individual segment matches up to the next present colon. The third
--- match (height) and everything thereafter needs to check up-to the next
--- colon -or- ending bracket since they could be the final segment.
---
--- Optional segments should of course have the "?" modifer attached to
--- their preceeding colon, and should use * for the content match rather
--- than +.
local IMAGE_PATTERN = [[{img%:([^:]+)%:([^:]+)%:([^:}]+)%:?([^:}]*)%}]];

--- Note that the image tag has to be outside a <P> tag.
local IMAGE_TAG = [[</P><img src="%s" width="%s" height="%s" align="%s"/><P>]];


-- This pattern matches all individual characters to replace. Each character
-- with a key in HTML_ESCAPE_REPLACEMENTS should be present here.
local HTML_ESCAPE_PATTERN = "[&<>\"]"

-- Table of replacements for characters to their HTML escaped equivalents.
local HTML_ESCAPE_REPLACEMENTS = {
    ["&"] = "&amp;",
    ["<"] = "&lt;",
    [">"] = "&gt;",
    ["\""] = "&quot;",
}

-- Convert alignment in tags to HTML friendly variant.
local alignmentTagConversion = {
	["c"] = "center",
	["l"] = "left",
	["r"] = "right",
};

-- Extract header # and alignment from tag and convert to HTML friendly format.
local tagToHTML = {
	-- Headers
	["{h(%d)}"] = "<h%1>",
	["{h(%d):c}"] = "<h%1 align=\"center\">",
	["{h(%d):r}"] = "<h%1 align=\"right\">",
	["{/h(%d)}"] = "</h%1>",
	-- Paragraphs
	["{p}"] = "<P>",
	["{p:c}"] = "<P align=\"center\">",
	["{p:r}"] = "<P align=\"right\">",
	["{/p}"] = "</P>",
};

function mrp:ConvertStringToHTML(text) -- Somewhat based on Ellypse's design, with help from Meorawr, with some modifications. Thanks guys. The verbose comments are for me to look back on later.
	-- First, escape naughty characters thats players may input into the box.
	text = string.gsub(text, HTML_ESCAPE_PATTERN, HTML_ESCAPE_REPLACEMENTS);
	
	-- Replace tags with HTML throughout the entire text. {h1:c} to <h1 align="center"> for example.
	for pattern, replacement in pairs(tagToHTML) do
		text = text:gsub(pattern, replacement);
	end

	local chunks = {}; -- Contains broken up segments of the profile, each key is a pair of tags and the text within them. It also contains areas between tags that aren't wrapped in tags, that we'll need to place tags around later.
	
	local i = 1;
	
	while text:find("<") and i < 500 do -- Loop through entire text up to 500 times and find each "<".
		local textBeforeFirstTag;
		textBeforeFirstTag = text:sub(1, text:find("<") - 1); -- textBeforeFirstTag will always be everything before the first "<" found, after bits are cut off of the beginning with each pass.
		if(#textBeforeFirstTag > 0) then
			table.insert(chunks, textBeforeFirstTag) -- Drop everything before the first "<" on this pass into the chunks table.
		end
		
		local fullTagString; -- This will contain entire lines wrapped in tags, including the tags themselves.
		
		local endTag = text:match("</(.-)>"); -- Look for next end tag and set this variable to it.
		
		if(endTag) then -- If there is an end tag..
			fullTagString = text:sub( text:find("<"), text:find("</") + #endTag + 2); -- Set fullTagString to text wrapped in tags, including tags themselves. 2 accounts for the / and the >.
			if(fullTagString == #endTag + 3) then -- If number of the full string is equal to the entire end tag, bail. I guess something went wrong.
				return
			end
			tinsert(chunks, fullTagString) -- Insert the text wrapped in tags into the chunks table.
		else
			return -- If no more end tags, return?
		end
		
		local remainingText; -- Remaining text will be what's left after we chop off what was before the tagged text, and the tagged text itself.
		remainingText = text:sub(#textBeforeFirstTag + #fullTagString + 1) -- Take only what's left after what we've already dealt with.
		text = remainingText;
		
		i = (i + 1);
	end
	
	if(#text > 0) then
		table.insert(chunks, text); -- Whatevers left must not be in any tags, near the end of the text. Stuff it into the chunks table.
	end
	
	local finalText = "";
	for _, chunk in pairs(chunks) do -- For every chunk we saved above do..
		if not chunk:find("<") then -- If a line doesn't have tags around it, it's invalid in the HTML frame. We put <P> tags around it so it's readable.
			chunk = "<P>" .. chunk .. "</P>";
		end
		chunk = chunk:gsub("\n", "<br/>"); -- Replace newlines with proper HTML line breaks.
		-- TRP IMAGE SUPPORT --
		-- Since we're using TRP's tag style here, we'll work with what they've got setup as a standard for now, unless we come up with something better later as a team.
		chunk = chunk:gsub(IMAGE_PATTERN, function(image, width, height, alignment) -- It looks like we're testing this pattern on the chunk to see if it's an image.
			-- If no alignment is provided or invalid, "center" will be returned.
			alignment = alignmentTagConversion[alignment] or "center";
			
			-- Prevent catastrophic meltdown by assigning 128 as width and height if these values come in non-numeric.
			width = tonumber(width) or 128;
			height = tonumber(height) or 128;
			
			-- Apparently negative inputs here cause issues, so we math.abs them to prevent that.
			return string.format(IMAGE_TAG, image, math.abs(width), math.abs(height), alignment);
		end);
		
		chunk = chunk:gsub("%!%[(.-)%]%((.-)%)", function(icon, size)
			if(icon:find("\\")) then -- "\\ means we found an icon path."
				local width, height;
				if size:find("%,") then
					width, height = strsplit(",", size);
				else
					width = tonumber(size) or 128;
					height = width;
				end
				-- Again, width and height must be absolute to prevent catastrophic meltdown.
				return string.format(IMAGE_TAG, icon, math.abs(width), math.abs(height), "center");
			end
			return mrp:ReturnTextureTag(icon, tonumber(size) or 25);
		end);
		
		chunk = chunk:gsub("%[(.-)%]%((.-)%)",
			"<a href=\"%2\">[%1]|r</a>");
		
		chunk = chunk:gsub("{link%*(.-)%*(.-)}",
			"<a href=\"%1\">[%2]|r</a>");

		finalText = finalText .. chunk; -- Finished text, add chunk to end of it.
	end
	
	return "<HTML><BODY>" .. finalText .. "</BODY></HTML>"; -- Wrap the entire text in <HTML> and <BODY> tags so the whole bloody thing works.
end